<?php

namespace BS\RealTimeChat\Pub\Controller;

use BS\RealTimeChat\Entity\Message;
use BS\RealTimeChat\Listener;
use BS\RealTimeChat\LongPoll\Communicator;
use XF\Mvc\ParameterBag;
use XF\Pub\Controller\AbstractController;

class Chat extends AbstractController
{
    protected function preDispatchController($action, ParameterBag $params)
    {
        if (!\XF::visitor()->canViewChat())
        {
            throw $this->exception($this->noPermission());
        }
    }

    public function actionIndex()
    {
        $visitor = \XF::visitor();

        $permissions = [
            'canDelete' => $visitor->hasChatPermission('canDelete'),
            'canEdit' => $visitor->hasChatPermission('canEdit'),
            'canEditAny' => $visitor->hasChatPermission('canEditAny'),
            'canDeleteAny' => $visitor->hasChatPermission('canDeleteAny'),
            'canReport' => $visitor->canReport()
        ];

        $viewParams = [
            'messages' => $this->getMessageRepo()->getMessagesForList(),

            'permissions' => $permissions
        ];

        return $this->view('BS\RealTimeChat:Full', 'real_time_chat', $viewParams);
    }

    protected function setupMessage()
    {
        $visitor = \XF::visitor();
        $message = $this->filter('message', 'str');

        $messageCreator = $this->service('BS\RealTimeChat:Message\Creator', $visitor);
        $messageCreator->setMessageContent($message);

        return $messageCreator;
    }

    protected function afterPost(\BS\RealTimeChat\Service\Message\Creator $messageCreator)
    {
        $messageCreator->sendNotifications();
        $message = $messageCreator->getMessage();
        $emitter = $this->app['chatEmitter'];

        if ($message->pm_user_id)
        {
            $emitter->to($message->user_id)
                ->to($message->pm_user_id)
                ->emit('new message', $message->forTemplate());
        }
        else
        {
            $emitter->emit('new message', $message->forTemplate());
        }
    }

    public function actionPost()
    {
        $this->assertXhrOnly();

        if (!\XF::visitor()->canChatWrite())
        {
            return $this->noPermission();
        }

        if ($this->filter('message_id', 'uint'))
        {
            return $this->rerouteController(__CLASS__, 'save');
        }

        $this->assertNotFlooding('chatMessage', $this->options()->realTimeChatSendTimeout);

        $messageCreator = $this->setupMessage();
        if (!$messageCreator->validate($errors))
        {
            return $this->error($errors);
        }

        $message = $messageCreator->save();

        if ($message)
        {
            $this->afterPost($messageCreator);
        }
        
        return $this->redirect($this->getDynamicRedirect());
    }

    protected function afterDelete(\BS\RealTimeChat\Entity\Message $message)
    {
        $emitter = $this->app['chatEmitter'];

        if ($message->pm_user_id)
        {
            $emitter->to($message->user_id)->to($message->pm_user_id)->emit('delete message', [
                'message_ids' => (array)$message->message_id
            ]);
        }
        else
        {
            $emitter->emit('delete message', [
                'message_ids' => (array)$message->message_id
            ]);
        }
    }

    public function actionDelete()
    {
        $this->assertXhrOnly();

        $message = $this->assertMessageExists($this->filter('message_id', 'uint'));

        if (!$message->canDelete())
        {
            return $this->noPermission();
        }

        $message->delete();

        $this->afterDelete($message);

        return $this->redirect($this->getDynamicRedirect());
    }

    protected function setupEdit(\BS\RealTimeChat\Entity\Message $message)
    {
        $visitor = \XF::visitor();
        $messageContent = $this->filter('message', 'str');

        $messageEditor = $this->service('BS\RealTimeChat:Message\Editor', $message, $visitor);
        $messageEditor->setMessageContent($messageContent);

        return $messageEditor;
    }

    protected function afterSave(\BS\RealTimeChat\Service\Message\Editor $editor)
    {
        $message = $editor->getMessage();
        $emitter = $this->app['chatEmitter'];

        if ($message->pm_user_id)
        {
            $emitter->to($message->pm_user_id)
                ->to($message->user_id)
                ->emit('edit message', [
                    'message_id' => $message->message_id,
                    'message' => $message->message_body
                ]);
        }
        else
        {
            $emitter->emit('edit message', [
                'message_id' => $message->message_id,
                'message' => $message->message_body
            ]);
        }
    }

    public function actionSave()
    {
        $this->assertXhrOnly();

        $message = $this->assertMessageExists($this->filter('message_id', 'uint'));

        if (!$message->canEdit())
        {
            return $this->noPermission();
        }

        $messageEditor = $this->setupEdit($message);
        if (!$messageEditor->validate($errors))
        {
            return $this->error($errors);
        }

        $messageEditor->save();

        $this->afterSave($messageEditor);

        return $this->redirect($this->getDynamicRedirect());
    }

    public function actionEdit()
    {
        $this->assertXhrOnly();

        $message = $this->assertMessageExists($this->filter('message_id', 'uint'));

        if (!$message->canEdit())
        {
            return $this->noPermission();
        }

        $viewParams = [
            'message' => [
                'message_id' => $message->message_id,
                'message' => $message->message
            ]
        ];

        return $this->view('BS\RealTimeChat:Json', '', $viewParams);
    }

    public function actionReport()
    {
        $message = $this->assertMessageExists($this->filter('message_id', 'uint'));
        if (!$message->canReport($error))
        {
            return $this->noPermission($error);
        }

        $reportPlugin = $this->plugin('XF:Report');
        return $reportPlugin->actionReport(
            'chat_message', $message,
            $this->buildLink('chat/report', null, ['message_id' => $message->message_id]),
            $this->getDynamicRedirect()
        );
    }

    protected function onVisitorWriting()
    {
        $visitor = \XF::visitor();

        $emitter = $this->app['chatEmitter'];
        $emitter->emit('user writing', [
            'user_id' => $visitor->user_id,
            'username' => $visitor->username
        ]);
    }

    public function actionWriting()
    {
        $this->onVisitorWriting();

        return $this->redirect($this->getDynamicRedirect());
    }

    public function actionLongPolling()
    {
        if ($this->options()->realTimeChatClientCommunication['communication'] != 'long_polling')
        {
            return $this->noPermission();
        }

        /** @var Communicator $communicator */
        $communicator = $this->app['longPollingCommunicator'];

        $view = $this->view();

        $events = $communicator->listen($this->filter('last_event_time', 'int'));

        if ($events && $events->count())
        {
            $view->setJsonParams([
                'events' => $events->toApiResults()->render(),
                'last_event_time' => $events->last()['event_time']
            ]);
        }

        return $view;
    }

    protected function userBanSaveProcess(\XF\Entity\User $user)
    {
        $form = $this->formAction();

        $input = $this->filter([
            'ban_length' => 'str',
            'unban_date' => 'datetime',
            'reason' => 'str',
            'length_value' => 'uint',
            'length_unit' => 'str'
        ]);

        switch ($input['ban_length']) 
        {
            case 'permanent':
                $input['unban_date'] = 0;
                break;
            
            case 'while':
                $input['unban_date'] = strtotime('+' . $input['length_value'] . ' ' . $input['length_unit']);
                break;
        }

        unset($input['ban_length'], $input['length_unit'], $input['length_value']);

        $banUser = $user->isBannedInChat() ? $user->ChatBan : $this->em()->create('BS\RealTimeChat:Banned'); 

        $form->setupEntityInput($banUser, $input);

        $form->setup(function() use ($banUser, $user)
        {
            $banUser->user_id = $user->user_id;
        });

        $form->saveEntity($banUser);

        return $form;
    }

    protected function afterBan(\XF\Entity\User $user)
    {
        $emitter = $this->app['chatEmitter'];
        $emitter->in($user->user_id)->emit('ban');
    }

    public function actionBan()
    {
        $visitor = \XF::visitor();

        if (!$visitor->hasChatPermission('canBan'))
        {
            return $this->noPermission();
        }

        if ($this->isPost())
        {
            $user = $this->em()->findOne('XF:User', ['username' => $this->filter('username', 'str')]);
            if (!$user)
            {
                return $this->notFound(\XF::phrase('requested_user_not_found'));
            }

            if ($this->filter('lift', 'bool'))
            {
                if (!$user->isBannedInChat())
                {
                    return $this->noPermission();
                }

                $user->ChatBan->delete();
                $this->afterBan($user);
            }
            else
            {
                if (!$visitor->canChatBan($user))
                {
                    return $this->noPermission();
                }

                $this->userBanSaveProcess($user)->run();
                $this->afterBan($user);
            }

            return $this->redirect($this->getDynamicRedirect());
        }

        $viewParams = [
            'username' => $this->filter('username', 'str')
        ];

        return $this->view('BS\RealTimeChat:Ban', 'real_time_chat_ban', $viewParams);
    }

    public function actionBanList()
    {
        if (!\XF::visitor()->canChatViewBanList())
        {
            return $this->noPermission();
        }

        $page = $this->filterPage();
        $perPage = 50;

        $banned = $this->finder('BS\RealTimeChat:Banned')->limitByPage($page, $perPage, 1)->fetch();

        if (!count($banned))
        {
            return $this->message(\XF::phrase('real_time_chat_banned_users_is_empty'));
        }

        $hasNext = count($banned) > $perPage;
        $banned = $banned->slice(0, $perPage);

        $viewParams = [
            'banned' => $banned,
            'hasNext' => $hasNext,
            'page' => $page
        ];

        return $this->view('BS\RealTimeChat:Ban\List', 'real_time_chat_ban_list', $viewParams);
    }

    public function actionHistory(ParameterBag $params)
    {
        if (!\XF::visitor()->canViewChatHistory())
        {
            return $this->noPermission();
        }

        if ($params->message_id)
        {
            return $this->rerouteController(__CLASS__, 'history-view', $params);
        }

        $messageFinder = $this->getMessageRepo()->findMessages();

        $page = $this->filterPage();
        $perPage = $this->options()->realTimeChatMessageLimit;

        $total = $messageFinder->total();

        $this->assertValidPage($page, $perPage, $total, 'chat/history');

        $messages = $messageFinder->limitByPage($page, $perPage)->fetch();

        $viewParams = [
            'messages' => $messages,

            'page' => $page,
            'perPage' => $perPage,
            'total' => $total
        ];

        return $this->view('BS\RealTimeChat:History\List', 'real_time_chat_history_list', $viewParams);
    }

    public function actionHistoryView(ParameterBag $params)
    {
        if (!\XF::visitor()->canViewChatHistory())
        {
            return $this->noPermission();
        }
        
        $message = $this->assertMessageExists($params->message_id);

        if (! $message->canView())
        {
            return $this->noPermission();
        }

        $viewParams = [
            'message' => $message
        ];

        return $this->view('BS\RealTimeChat:History\View', 'real_time_chat_history_view', $viewParams);
    }

    public function assertXhrOnly()
    {
        if (!$this->request->isXhr())
        {
            throw $this->exception($this->error(
                \XF::phrase('rtc_action_available_via_xhr_only'), 405
            ));
        }
    }

    /**
     * @return \BS\RealTimeChat\Repository\Message
     */
    protected function getMessageRepo()
    {
        return $this->repository('BS\RealTimeChat:Message');
    }

    /**
     * @param $id
     * @param null $with
     * @return Message
     * @throws \XF\Mvc\Reply\Exception
     */
    protected function assertMessageExists($id, $with = null)
    {
        return $this->assertRecordExists('BS\RealTimeChat:Message', $id, $with, 'real_time_chat_requested_message_not_found');
    }
}